#include <stdlib.h>
#include <iostream>
#include <string>
#include <vector>
#include <fstream>
#include <sstream>
#include <cstring>

using std::string;
using std::getline;
using std::vector;
using std::fstream;

#define BEARLIB_HS_DLL
#include "bearlibhs.h"

// Implementation
unsigned int fnv_hash ( void *key, int len )
{
   unsigned char *p = (unsigned char*)key;
   unsigned int h = 2166136261;
   int i;

   for ( i = 0; i < len; i++ )
     h = ( h * 16777619 ) ^ p[i];

   return h;
}

string              BearHS_filename;
vector<HS_Entry>    BearHS_entries;

// Ripped somewhere on the internet, nice function
template <class T> bool from_string
    ( T& t, const std::string& s, std::ios_base& (*f)(std::ios_base&) )
{
  std::istringstream iss(s);
  return !(iss >> f >> t).fail();
}

string trim( string str )
{
    // str already a copy, ok to modify

    str.erase(str.find_last_not_of(' ')+1);
    str.erase(0, str.find_first_not_of(' '));
	return str;
}

void BearHS__PadNames()
{
    // Pad strings with zeroes to get correct hash
	for ( size_t i=0; i<BearHS_entries.size(); i++ )
	{
		char* name = BearHS_entries[i].name;
		for ( int i=strlen(name); i<NAME_LENGTH; i++ ) name[i] = '\0';
	}
}

bool BearHS_UseFile__ReadFile()
{
	fstream stream( BearHS_filename.c_str(), std::ios_base::in );
	if ( stream.fail() ) return false; // Cannot open, doesn't exist? Okay
	
	// Header
	
	string header; getline( stream, header );
	if ( stream.eof() ) return false; // Score file of header only is useless
	
	size_t pos;
	pos = header.find_first_of( ':' );
	if ( pos == string::npos ) return false; // Broken header
	string hash_str = header.substr( pos+2 /* + space after ':' */ );
	unsigned int hash = 0;
	if ( !from_string<unsigned int>( hash, hash_str, std::hex ) ) return false;
	
	// Entries
	
	int read = 0;
	
	while ( !stream.eof() )
	{
		string line; getline( stream, line );
		if ( line.length() == 0 ) break;
		
		pos = line.find_first_of( '=' );
		if ( pos == string::npos ) return false; // Broken entry
		
		string name = trim(line.substr( 0, pos ));
		string score = line.substr( pos+1 );

        HS_Entry& entry = BearHS_entries[read];
		if ( !from_string<int>( entry.score, score, std::dec ) ) return false;
		strncpy( entry.name, name.c_str(), NAME_LENGTH );
        entry.name[name.length() >= NAME_LENGTH ? NAME_LENGTH-1 : name.length()] = '\0';
		
		if ( ++read == BearHS_entries.size() ) break;
	}

	stream.close();
	
	// Check hash

	BearHS__PadNames();
	
	HS_Entry* array = &BearHS_entries.front();
	unsigned int correct_hash =
		fnv_hash( (void*)array, sizeof(HS_Entry)*BearHS_entries.size() );
		
	if ( hash != correct_hash )
	{
		return false;
	}
	
	return true;
}

void HS_Clear__Nullify( int number_of_entries )
{
    BearHS_entries.clear();
	BearHS_entries.resize( number_of_entries );
    for ( int i=0; i<number_of_entries; i++ )
    {
        BearHS_entries[i].name[0] = '\0';
        BearHS_entries[i].score = 0;
    }
}

void BEARLIB_API HS_Clear()
{
    HS_Clear__Nullify( BearHS_entries.size() );
}

void BEARLIB_API HS_UseFile( const char* filename, unsigned int number_of_entries )
{
	BearHS_filename = string( filename );

	if ( number_of_entries < 1 ) number_of_entries = 1;
	BearHS_entries.resize( number_of_entries );

	if ( !BearHS_UseFile__ReadFile() )
	{
		// Cannot open or read file, revert to empty score table
		HS_Clear__Nullify( number_of_entries );
	}
}

void BEARLIB_API HS_Post( const char* name, int score )
{
    if ( strlen(name) == 0 )
	{
		// Wrong
		return;
	}

	int size = BearHS_entries.size();
	
	if ( score < BearHS_entries[size-1].score )
	{
		// Nothing to do
		return;
	}
	
	int index = size-1;
	
	BearHS_entries[index].score = score;
	strncpy( BearHS_entries[index].name, name, NAME_LENGTH );
	
	if ( strlen(name) >= NAME_LENGTH )
        BearHS_entries[index].name[NAME_LENGTH-1] = '\0';
	
	// Float it up
	while ( index > 0 && (score > BearHS_entries[index-1].score) )
	{
		HS_Entry temp = BearHS_entries[index-1];
		BearHS_entries[index-1] = BearHS_entries[index];
		BearHS_entries[index] = temp;
		
		index -= 1;
	}
	
    BearHS__PadNames();
	
	HS_Entry* array = &BearHS_entries.front();
	unsigned int hash =
		fnv_hash( (void*)array, sizeof(HS_Entry)*BearHS_entries.size() );
	
	std::fstream stream( BearHS_filename.c_str(), std::ios_base::out );
	if ( stream.fail() ) return; // Oh, whatever
	
	stream << "# HASH: " << std::hex << hash << std::dec << std::endl;
	for ( size_t i=0; i<BearHS_entries.size(); i++ )
	{
		if ( strlen( BearHS_entries[i].name ) == 0 ) break;
		
		stream << BearHS_entries[i].name << " = ";
		stream << BearHS_entries[i].score << std::endl;
	}
	stream.close();
}

HS_Entry BEARLIB_API HS_Get( int entry_index )
{
	if ( entry_index < 0 || (size_t)entry_index >= BearHS_entries.size() )
	{
        HS_Entry result = { "", 0 };
		return result;
	}
	
	return BearHS_entries[entry_index];
}

int BEARLIB_API HS_Count()
{
	int index = BearHS_entries.size()-1;
	for ( ; index >= 0 && BearHS_entries[index].name[0] == '\0'; index-- );
	return (index+1);
}